// dependance of context.h
#include <gcc-plugin.h>
// 'g' global variable
#include <context.h>
// dependance of gimple.h
#include <tree.h>
// dependance of gimple-iterator.h
#include <gimple.h>
// gsi
#include <gimple-iterator.h>
// basename
#include <filesystem>
// set
#include <set>

// our pass
#include "pass_mpi_collective.hpp"

// Definitions
const pass_data my_pass_data =
{
	GIMPLE_PASS, /* type */
	"MPI COLLECTIVE", /* name */
	OPTGROUP_NONE, /* optinfo_flags */
	TV_OPTIMIZE, /* tv_id */
	0, /* properties_required */
	0, /* properties_provided */
	0, /* properties_destroyed */
	0, /* todo_flags_start */
	0, /* todo_flags_finish */
};

// Class functions
pass_mpi_collective::pass_mpi_collective(gcc::context *ctxt)
	: opt_pass(my_pass_data, ctxt) {}

pass_mpi_collective *pass_mpi_collective::clone()
{
	return new pass_mpi_collective(g);
}

bool pass_mpi_collective::gate(function *fun)
{
	//printf("In gate of: %s\n", fndecl_name(fun->decl));
	printf("In gate of: %s\n", function_name(fun));
	//printf("In gate of: %s\n", current_function_name());
	return true;
}

unsigned int pass_mpi_collective::execute(function *fun)
{
	printf("In execute of: %s\n", function_name(fun));

	print_tree(fun);
	split_blocks(fun);
	printf("\t--------------------------[split]---------------------------\n");
	print_tree(fun);

	alloc_bb_aux(fun);
	label_collec(fun);
	label_dom(fun);
	label_dom_front(fun);

	alloc_edge_aux(fun);

	mark_edge(fun);
	reset_bb_mark(fun);

	cfgviz_dump(fun, "_split");

	free_dom_data();
	free_bb_aux(fun);
	free_edge_aux(fun);

	return 0;
}

void pass_mpi_collective::print_tree(function *fun)
{
	basic_block bb;
	gimple_stmt_iterator gsi;
	gimple *stmt;
	//size_t line;

	FOR_EACH_BB_FN(bb, fun)
	{
		printf("\tBasic block: %d\n", bb->index);
		for (gsi = gsi_start_bb(bb); !gsi_end_p(gsi); gsi_next(&gsi))
		{
			stmt = gsi_stmt(gsi);
			//line = gimple_lineno(stmt);
			//printf("\tstatement at line = %lu\n", line);
			if (is_func(stmt))
			{
				printf("\t\tStatement is a function call (%s)\n",
				       fndecl_name(gimple_call_fndecl(stmt)));
				if (__is_mpi_func(stmt))
				{
					printf("\t\t\tStatement is a MPI function call (%s)\n",
					       fndecl_name(gimple_call_fndecl(stmt)));
					if (__is_mpi_collec(stmt) != LAST_AND_UNUSED_MPI_COLLECTIVE_CODE)
					{
						printf("\t\t\t\tStatement is a MPI collective (%s)\n",
						       fndecl_name(gimple_call_fndecl(stmt)));
					}
				}
			}
		}
	}
}

// MPI function / collectives detections

bool pass_mpi_collective::is_func(gimple *stmt)
{
	return is_gimple_call(stmt);
}

bool pass_mpi_collective::__is_mpi_func(gimple *stmt)
{
	return strncmp(fndecl_name(gimple_call_fndecl(stmt)), "MPI_", 4) == 0;
}

bool pass_mpi_collective::is_mpi_func(gimple *stmt)
{
	return is_func(stmt) && __is_mpi_func(stmt);
}

mpi_collective_code pass_mpi_collective::__is_mpi_collec(gimple *stmt)
{
	size_t i;
	const char *callee_name;

	i = 0;
	callee_name = fndecl_name(gimple_call_fndecl(stmt));
	while (i < LAST_AND_UNUSED_MPI_COLLECTIVE_CODE)
	{
		if (strcmp(mpi_collective_name[i], callee_name) == 0)
		{
			return (enum mpi_collective_code) i;
		}
		++i;
	}
	return LAST_AND_UNUSED_MPI_COLLECTIVE_CODE;
}

mpi_collective_code pass_mpi_collective::is_mpi_collec(gimple *stmt)
{
	if (!is_mpi_func(stmt))
	{
		return LAST_AND_UNUSED_MPI_COLLECTIVE_CODE;
	}
	return __is_mpi_collec(stmt);
}

// split blocks according to mpi collectives

void pass_mpi_collective::split_blocks(function *fun)
{
	basic_block bb;
	gimple_stmt_iterator gsi;
	gimple *stmt;
	gimple *prev_stmt;
	size_t nb_collective;
	bool split;

	FOR_EACH_BB_FN(bb, fun)
	{
		nb_collective = 0;
		stmt = NULL;
		split = false;
		for (gsi = gsi_start_bb(bb); !gsi_end_p(gsi) && ! split; gsi_next(&gsi))
		{
			prev_stmt = stmt;
			stmt = gsi_stmt(gsi);
			if (is_mpi_collec(stmt) != LAST_AND_UNUSED_MPI_COLLECTIVE_CODE)
			{
				++nb_collective;
			}
			if (nb_collective >= 2)
			{
				split_block(bb, prev_stmt);
				split = true;
				//The inner for doesn't stop naturally, whereas the
				//FOR_EACH_BB_FN *will* iterate on the new bb we just split.
				//hence the split=true
			}
		}
	}
}

void pass_mpi_collective::alloc_bb_aux(function *fun)
{
	basic_block bb;
	FOR_ALL_BB_FN(bb, fun)
	{
		bb->aux = (bb_data *) new bb_data();
	}
}

// gestions des aux
void pass_mpi_collective::label_collec(function *fun)
{
	basic_block bb;
	gimple_stmt_iterator gsi;
	gimple *stmt;
	mpi_collective_code mpi_code;

	FOR_ALL_BB_FN(bb, fun)
	{
		((bb_data *) bb->aux)->mpi_code = LAST_AND_UNUSED_MPI_COLLECTIVE_CODE;
		for (gsi = gsi_start_bb(bb); !gsi_end_p(gsi); gsi_next(&gsi))
		{
			stmt = gsi_stmt(gsi);
			mpi_code = is_mpi_collec(stmt);
			if (mpi_code != LAST_AND_UNUSED_MPI_COLLECTIVE_CODE)
			{
				((bb_data *) bb->aux)->mpi_code = mpi_code;
			}
		}
	}
}

void pass_mpi_collective::free_bb_aux(function *fun)
{
	basic_block bb;
	FOR_ALL_BB_FN(bb, fun)
	{
		if (bb->aux != NULL)
		{
			delete (bb_data *) bb->aux;
			bb->aux = NULL;
		}
	}
}

// calculate dominance
void pass_mpi_collective::label_dom(function *fun)
{
	basic_block bb;
	auto_vec<basic_block> dom_gccvec;
	auto_vec<basic_block> post_dom_gccvec;
	size_t size_dom;
	size_t size_post_dom;

	calculate_dominance_info(CDI_DOMINATORS);
	calculate_dominance_info(CDI_POST_DOMINATORS);

	FOR_ALL_BB_FN(bb, fun)
	{
		bb_data *data = ((bb_data *) bb->aux);

		dom_gccvec = get_all_dominated_blocks(CDI_DOMINATORS, bb);
		size_dom = dom_gccvec.length();
		for (size_t i = 0; i < size_dom; ++i)
		{
			if (dom_gccvec[i]->index != bb->index)
			{
				data->dom.insert(dom_gccvec[i]);
			}
		}

		//for (basic_block bb2 : *dom)
		//{
		//	printf("%d dom %d\n", bb->index, bb2->index);
		//}
		//((bb_data *) bb->aux)->dom = dominated;

		post_dom_gccvec = get_all_dominated_blocks(CDI_POST_DOMINATORS, bb);
		size_post_dom = post_dom_gccvec.length();
		for (size_t i = 0; i < size_post_dom; ++i)
		{
			if (post_dom_gccvec[i]->index != bb->index)
			{
				data->post_dom.insert(post_dom_gccvec[i]);
			}
		}
		//for (basic_block bb2 : *post_dom)
		//{
		//	printf("%d post dom %d\n", bb->index, bb2->index);
		//}
		//((bb_data *) bb->aux)->post_dom = post_dominated;
	}
}

void pass_mpi_collective::label_dom_front(function *fun)
{
	edge p;
	edge_iterator ei;
	basic_block bb;
	basic_block dom_bb;
	basic_block runner;
	bb_data *data;

	FOR_EACH_BB_FN(bb, fun)
	{
		if (EDGE_COUNT(bb->preds) < 2)
		{
			continue;
		}
		dom_bb = get_immediate_dominator(CDI_DOMINATORS, bb);
		FOR_EACH_EDGE(p, ei, bb->preds)
		{
			runner = p->src;
			if (runner == ENTRY_BLOCK_PTR_FOR_FN(fun))
			{
				continue;
			}
			while (runner != dom_bb)
			{
				data = (bb_data *) runner->aux;
				if (data->dom_front.count(bb) > 0)
				{
					break;
				}
				data->dom_front.insert(bb);
				runner = get_immediate_dominator(CDI_DOMINATORS, runner);
			}
		}
	}

	FOR_EACH_BB_FN(bb, fun)
	{
		if (EDGE_COUNT(bb->succs) < 2)
		{
			continue;
		}
		dom_bb = get_immediate_dominator(CDI_POST_DOMINATORS, bb);
		FOR_EACH_EDGE(p, ei, bb->succs)
		{
			runner = p->dest;
			if (runner == EXIT_BLOCK_PTR_FOR_FN(fun))
			{
				continue;
			}
			while (runner != dom_bb)
			{
				data = (bb_data *) runner->aux;
				if (data->post_dom_front.count(bb) > 0)
				{
					break;
				}
				data->post_dom_front.insert(bb);
				runner = get_immediate_dominator(CDI_POST_DOMINATORS, runner);
			}
		}
	}
}

void pass_mpi_collective::free_dom_data()
{
	free_dominance_info(CDI_DOMINATORS);
	free_dominance_info(CDI_POST_DOMINATORS);
}

// parkour de graph
void pass_mpi_collective::reset_bb_mark(function *fun)
{
	basic_block bb;

	FOR_ALL_BB_FN(bb, fun)
	{
		((bb_data *) bb->aux)->mark = 0;
	}
}

void pass_mpi_collective::alloc_edge_aux(function *fun)
{
	basic_block bb;
	edge e;
	edge_iterator ei;

	FOR_ALL_BB_FN(bb, fun)
	{
		FOR_EACH_EDGE(e, ei, bb->succs)
		{
			e->aux = new edge_data();
		}
	}
}

void pass_mpi_collective::free_edge_aux(function *fun)
{
	basic_block bb;
	edge e;
	edge_iterator ei;

	FOR_ALL_BB_FN(bb, fun)
	{
		FOR_EACH_EDGE(e, ei, bb->succs)
		{
			if (e->aux != NULL)
			{
				delete (edge_data *) e->aux;
				e->aux = NULL;
			}
		}
	}
}

void pass_mpi_collective::mark_edge(function *fun)
{
	mark_edge(ENTRY_BLOCK_PTR_FOR_FN(fun), 1);
}

void pass_mpi_collective::mark_edge(basic_block bb, int mark)
{
	edge e;
	edge_iterator ei;
	basic_block bb_dest;
	int dest_mark;

	((bb_data *) bb->aux)->mark = mark;
	FOR_EACH_EDGE(e, ei, bb->succs)
	{
		bb_dest = e->dest;
		dest_mark = ((bb_data *) bb_dest->aux)->mark;
		if (dest_mark > 0 && dest_mark < mark)
		{
			((edge_data *) e->aux)->loop = true;
		}
	}

	FOR_EACH_EDGE(e, ei, bb->succs)
	{
		bb_dest = e->dest;
		if (((bb_data *) bb_dest->aux)->mark == 0)
		{
			mark_edge(bb_dest, mark + 1);
		}
	}
}

//============================	GRAPHVIZ  =====================================

/* Build a filename (as a string) based on function name */
char *pass_mpi_collective::cfgviz_generate_filename(function *fun,
  const char *suffix)
{
	char *target_filename;
	int line;

	target_filename = (char *)xmalloc(1024 * sizeof(char));
	line = LOCATION_LINE(fun->function_start_locus);
	const char *mfilename =
	  std::filesystem::path(LOCATION_FILE(fun->function_start_locus))
	  .filename().c_str();

	snprintf(target_filename, 1024, "%s%s_%s_%d_%s.dot",
	         "dot/",
	         mfilename,
	         function_name(fun),
	         line,
	         suffix);

	return target_filename;
}

/* Dump the graphviz representation of function 'fun' in file 'out' */
void pass_mpi_collective::cfgviz_internal_dump(function *fun, FILE *out)
{
	basic_block bb;
	edge_iterator eit;
	edge e;

	fprintf(out, "Digraph G{\n");
	FOR_ALL_BB_FN(bb, fun)
	{
		bb_data *data = ((bb_data *) bb->aux);

		fprintf(out, "%d [label=\"BB %d", bb->index, bb->index);
		if (data->mpi_code != LAST_AND_UNUSED_MPI_COLLECTIVE_CODE)
		{

			fprintf(out, "|%s", mpi_collective_name[data->mpi_code]);
		}

		if (! data->dom.empty())
		{
			fprintf(out, "|dom(");
			for (basic_block bb2 : data->dom)
			{
				fprintf(out, "%d,", bb2->index);
			}
			fprintf(out, ")");
		}

		if (! data->post_dom.empty())
		{
			fprintf(out, "|post_dom(");
			for (basic_block bb2 : data->post_dom)
			{
				fprintf(out, "%d,", bb2->index);
			}
			fprintf(out, ")");
		}

		if (! data->dom_front.empty())
		{
			fprintf(out, "|dom_front(");
			for (basic_block bb2 : data->dom_front)
			{
				fprintf(out, "%d,", bb2->index);
			}
			fprintf(out, ")");
		}

		if (! data->post_dom_front.empty())
		{
			fprintf(out, "|post_dom_front(");
			for (basic_block bb2 : data->post_dom_front)
			{
				fprintf(out, "%d,", bb2->index);
			}
			fprintf(out, ")");
		}

		fprintf(out, "\" shape=ellipse]\n");

		FOR_EACH_EDGE(e, eit, bb->succs)
		{
			const char *label = "";
			if (e->flags == EDGE_TRUE_VALUE)
			{
				label = "true";
			}
			else if (e->flags == EDGE_FALSE_VALUE)
			{
				label = "false";
			}

			if (((edge_data *) e->aux)->loop)
			{
				label = "remove_loop";
			}

			fprintf(out, "%d -> %d [color=red label=\"%s\"]\n",
			        bb->index, e->dest->index, label);
		}
	}
	fprintf(out, "}\n");
}

void pass_mpi_collective::cfgviz_dump(function *fun, const char *suffix)
{
	char *target_filename;
	FILE *out;

	target_filename = cfgviz_generate_filename(fun, suffix);

	printf("[GRAPHVIZ] Generating CFG of function %s in file <%s>\n",
	       current_function_name(), target_filename);

	out = fopen(target_filename, "w");

	cfgviz_internal_dump(fun, out);

	fclose(out);
	free(target_filename);
}
